/***************************************************************************
                          bowl.cpp  -  description
                             -------------------
    begin                : Wed Mar 1 2000
    copyright            : (C) 2000 by Michael Speck
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "bowl.h"
#include "defs.h"
#include "soundserver.h"

Bowl::Bowl(int bx, int by, SDL_Surface *scr, SDLSurface *bkgnd, SDLSurface *bl, HiScoreEntry *p, int ln, LevelInit *level_init, AdvLevelInit *adv_li, int hd, ShrapnellList *s_list, SDL_Rect *rects, int *rect_num, int *tl, Controls *plyr, InitData *idata)
	 : screen(scr), backgnd(bkgnd), blocks(bl), player(p), level_num(ln), anim_l(idata->anim_l), trp(idata->transparency), ctrls(plyr), total_lines(tl), enemy(0), wholes(idata->wholes), send_all(idata->send_all), send_tetris(idata->send_tetris), smooth(idata->smooth_hori), help(idata->help), init_data(idata)
{
	//set draw rects
	Bowl::rects = rects;
	Bowl::rect_num = rect_num;

	//setup bowl
	bowl_x = bx;
	bowl_y = by;
	bowl_vis_w = BOWL_PIX_W;
	if (bowl_x < 0)
	    bowl_vis_w += bowl_x;
	if (bowl_x + BOWL_PIX_W > scr->w)
	    bowl_vis_w -= (BOWL_PIX_W + bowl_x) - scr->w;
	bowl_vis_h = BOWL_PIX_H;
	if (bowl_y < 0)
	    bowl_vis_h += bowl_y;
	if (bowl_y + BOWL_PIX_H > scr->h)
	    bowl_vis_h -= (BOWL_PIX_H + bowl_y) - scr->h;
	ClearBowl();
	last_ctrls = new Controls;
	last_ctrls->right = last_ctrls->left = last_ctrls->rot_left = last_ctrls->rot_right = 0;
	Bowl::level_init = level_init;
	adv_l_init = adv_li;
	new_next = 0;
    last_action = 0;
    rem_lines= 0;
    shrapnells = s_list;
    if (!smooth)
        hori_delay = 50 + hd * 10;
    else
        // hori delay now understood as velocity per second
        hori_delay =  (BLOCKSIZE + (10 - hd) * BLOCKSIZE) * 2;

    //draw bowl frame
    SD_CBx(backgnd->Surface()->pixels, backgnd->Width(), bowl_x, bowl_y, BOWL_PIX_W, BOWL_PIX_H/*bowl_vis_h*/, 0x0, BOWL_DARK);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x-3, bowl_y, 2, BOWL_PIX_H, BC_NRML);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x-2, bowl_y+1, 2, BOWL_PIX_H, BC_DARK);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x+BOWL_PIX_W, bowl_y, 2, BOWL_PIX_H, BC_NRML);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x+BOWL_PIX_W + 1, bowl_y+1, 2, BOWL_PIX_H, BC_DARK);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x-3, bowl_y + BOWL_PIX_H, BOWL_PIX_W + 5, 2, BC_NRML);
    SD_Rct(backgnd->Surface()->pixels, backgnd->Width(), bowl_x-2, bowl_y + BOWL_PIX_H + 1, BOWL_PIX_W + 4, 2, BC_DARK);
	
#ifdef SOUND
    //sounds
    snd_explosion = new Wave("explosion.wav");
    snd_rotate = new Wave("leftright.wav");
#endif

	//set line_off
	line_offset = level_init[player->level].lines;
	
	//setup block
	block = new Block((int*)bowl, smooth, idata->bl_by_bl);
	
	//init blocks
	block_num = 7;
	block_init = new BlockInit[block_num];
	
	block_init[0].rx = 0;
	block_init[0].ry = 0;
	block_init[0].id = 0;
	memset(block_init[0].blocks, 0, sizeof(block_init[0].blocks));
	block_init[0].blocks[0][1][1] = 1;
	block_init[0].blocks[0][2][1] = 1;
	block_init[0].blocks[0][1][2] = 1;
	block_init[0].blocks[0][2][2] = 1;
	block_init[0].blocks[1][1][1] = 1;
	block_init[0].blocks[1][2][1] = 1;
	block_init[0].blocks[1][1][2] = 1;
	block_init[0].blocks[1][2][2] = 1;
	block_init[0].blocks[2][1][1] = 1;
	block_init[0].blocks[2][2][1] = 1;
	block_init[0].blocks[2][1][2] = 1;
	block_init[0].blocks[2][2][2] = 1;
	block_init[0].blocks[3][1][1] = 1;
	block_init[0].blocks[3][2][1] = 1;
	block_init[0].blocks[3][1][2] = 1;
	block_init[0].blocks[3][2][2] = 1;
	
	block_init[1].rx = 0;
	block_init[1].ry = 0;
	block_init[1].id = 1;
	memset(block_init[1].blocks, 0, sizeof(block_init[1].blocks));
	block_init[1].blocks[0][0][2] = 1;
	block_init[1].blocks[0][1][2] = 1;
	block_init[1].blocks[0][2][2] = 1;
	block_init[1].blocks[0][3][2] = 1;
	block_init[1].blocks[1][1][0] = 1;
	block_init[1].blocks[1][1][1] = 1;
	block_init[1].blocks[1][1][2] = 1;
	block_init[1].blocks[1][1][3] = 1;
	block_init[1].blocks[2][0][2] = 1;
	block_init[1].blocks[2][1][2] = 1;
	block_init[1].blocks[2][2][2] = 1;
	block_init[1].blocks[2][3][2] = 1;
	block_init[1].blocks[3][1][0] = 1;
	block_init[1].blocks[3][1][1] = 1;
	block_init[1].blocks[3][1][2] = 1;
	block_init[1].blocks[3][1][3] = 1;
	
	block_init[2].rx = 0;
	block_init[2].ry = 1;
	block_init[2].id = 2;
	memset(block_init[2].blocks, 0, sizeof(block_init[2].blocks));
	block_init[2].blocks[0][0][1] = 1;
	block_init[2].blocks[0][1][1] = 1;
	block_init[2].blocks[0][2][1] = 1;
	block_init[2].blocks[0][1][2] = 1;
	block_init[2].blocks[1][1][0] = 1;
	block_init[2].blocks[1][1][1] = 1;
	block_init[2].blocks[1][1][2] = 1;
	block_init[2].blocks[1][0][1] = 1;
	block_init[2].blocks[2][0][1] = 1;
	block_init[2].blocks[2][1][0] = 1;
	block_init[2].blocks[2][1][1] = 1;
	block_init[2].blocks[2][2][1] = 1;
	block_init[2].blocks[3][1][0] = 1;
	block_init[2].blocks[3][1][1] = 1;
	block_init[2].blocks[3][1][2] = 1;
	block_init[2].blocks[3][2][1] = 1;
	
	block_init[3].rx = 0;
	block_init[3].ry = 1;
	block_init[3].id = 3;
	memset(block_init[3].blocks, 0, sizeof(block_init[3].blocks));
	block_init[3].blocks[0][0][2] = 1;
	block_init[3].blocks[0][1][1] = 1;
	block_init[3].blocks[0][1][2] = 1;
	block_init[3].blocks[0][2][1] = 1;
	block_init[3].blocks[1][0][1] = 1;
	block_init[3].blocks[1][0][2] = 1;
	block_init[3].blocks[1][1][2] = 1;
	block_init[3].blocks[1][1][3] = 1;
	block_init[3].blocks[2][0][2] = 1;
	block_init[3].blocks[2][1][1] = 1;
	block_init[3].blocks[2][1][2] = 1;
	block_init[3].blocks[2][2][1] = 1;
	block_init[3].blocks[3][0][1] = 1;
	block_init[3].blocks[3][0][2] = 1;
	block_init[3].blocks[3][1][2] = 1;
	block_init[3].blocks[3][1][3] = 1;
	
	block_init[4].rx = 0;
	block_init[4].ry = 1;
	block_init[4].id = 4;
	memset(block_init[4].blocks, 0, sizeof(block_init[4].blocks));
	block_init[4].blocks[0][0][1] = 1;
	block_init[4].blocks[0][1][1] = 1;
	block_init[4].blocks[0][1][2] = 1;
	block_init[4].blocks[0][2][2] = 1;
	block_init[4].blocks[1][1][1] = 1;
	block_init[4].blocks[1][1][2] = 1;
	block_init[4].blocks[1][0][2] = 1;
	block_init[4].blocks[1][0][3] = 1;
	block_init[4].blocks[2][0][1] = 1;
	block_init[4].blocks[2][1][1] = 1;
	block_init[4].blocks[2][1][2] = 1;
	block_init[4].blocks[2][2][2] = 1;
	block_init[4].blocks[3][1][1] = 1;
	block_init[4].blocks[3][1][2] = 1;
	block_init[4].blocks[3][0][2] = 1;
	block_init[4].blocks[3][0][3] = 1;
	
	block_init[5].rx = 0;
	block_init[5].ry = 1;
	block_init[5].id = 5;
	memset(block_init[5].blocks, 0, sizeof(block_init[5].blocks));
	block_init[5].blocks[0][0][1] = 1;
	block_init[5].blocks[0][1][1] = 1;
	block_init[5].blocks[0][2][1] = 1;
	block_init[5].blocks[0][0][2] = 1;
	block_init[5].blocks[1][0][0] = 1;
	block_init[5].blocks[1][1][0] = 1;
	block_init[5].blocks[1][1][1] = 1;
	block_init[5].blocks[1][1][2] = 1;
	block_init[5].blocks[2][0][1] = 1;
	block_init[5].blocks[2][1][1] = 1;
	block_init[5].blocks[2][2][1] = 1;
	block_init[5].blocks[2][2][0] = 1;
	block_init[5].blocks[3][1][0] = 1;
	block_init[5].blocks[3][1][1] = 1;
	block_init[5].blocks[3][1][2] = 1;
	block_init[5].blocks[3][2][2] = 1;
	
	block_init[6].rx = 0;
	block_init[6].ry = 1;
	block_init[6].id = 6;
	memset(block_init[6].blocks, 0, sizeof(block_init[6].blocks));
	block_init[6].blocks[0][2][2] = 1;
	block_init[6].blocks[0][0][1] = 1;
	block_init[6].blocks[0][1][1] = 1;
	block_init[6].blocks[0][2][1] = 1;
	block_init[6].blocks[1][0][2] = 1;
	block_init[6].blocks[1][1][0] = 1;
	block_init[6].blocks[1][1][1] = 1;
	block_init[6].blocks[1][1][2] = 1;
	block_init[6].blocks[2][0][0] = 1;
	block_init[6].blocks[2][0][1] = 1;
	block_init[6].blocks[2][1][1] = 1;
	block_init[6].blocks[2][2][1] = 1;
	block_init[6].blocks[3][1][0] = 1;
	block_init[6].blocks[3][1][1] = 1;
	block_init[6].blocks[3][1][2] = 1;
	block_init[6].blocks[3][2][0] = 1;
	
	// destination
	dest_x = dest_y = dest_state = 0;
}

Bowl::~Bowl()
{
	if (block_init) delete block_init;
	if (last_ctrls) delete last_ctrls;
#ifdef SOUND
	if (snd_explosion) delete snd_explosion;
	if (snd_rotate) delete snd_rotate;
#endif	
}

void Bowl::ClearBowl()
{
	int x, y;
	//left, right border
	for (y = 0; y < BOWL_PIX_H + 1; y++) {
		bowl[0][y] = 1;
		bowl[BOWL_W + 1][y] = 1;
	}
	//ground	
    for (x = 0; x < BOWL_W + 2; x++)
    	for (y = 0; y < BLOCKSIZE; y++)
    		bowl[x][BOWL_PIX_H + y] = 1;
	//interior
	for (x = 0; x < BOWL_W; x++)		
		for (y = 0; y < BOWL_PIX_H; y++)
			bowl[x + 1][y] = 0;
}

int Bowl::Update(int ms, char *keystate)
{
	int ret = BOWL_RET_NONE;
	if (keystate[ctrls->left]) {
		//left movement
		if (last_ctrls->left)
			block->Left(ms);
		else
			block->Left(-1); //direct movement
	}
	if (keystate[ctrls->right]) {
		//right movement
		if (last_ctrls->right)
			block->Right(ms);
		else
			block->Right(-1); //direct movement
	}
	if (keystate[ctrls->down]) {
		//down movement
		block->ForceSpeed(0.45); //drop with highest speed
	}
	if (keystate[ctrls->rot_left] && !last_ctrls->rot_left) {
		//left rotation
		block->RotLeft();
#ifdef SOUND
		sndsrv->PlayWave(snd_rotate);
#endif		
	}
	if (keystate[ctrls->rot_right] && !last_ctrls->rot_right) {
		//right rotation
		block->RotRight();
#ifdef SOUND		
		sndsrv->PlayWave(snd_rotate);
#endif
	}
	
	// fast drop
	if ( !keystate[ctrls->drop] )
	    block->ClearWaitDrop();

    if ( keystate[ctrls->drop] )
        block->SetDropFlag(1);
		
	if (*rect_num < DRAWRECTS) {
		SDL_Rect rect = {0, screen->h - 1, screen->w, 1};
		rects[(*rect_num)++] = rect;
	}
	
	last_ctrls->left = keystate[ctrls->left];
	last_ctrls->right = keystate[ctrls->right];
	last_ctrls->down = keystate[ctrls->down];
	last_ctrls->rot_left = keystate[ctrls->rot_left];
	last_ctrls->rot_right = keystate[ctrls->rot_right];
	if (adv_l_init != 0)
		last_action += ms;
	
	if (help)	
    	Dest_Hide();
	switch (block->Update(ms)) {
		case BLOCK_NOMOVE:
		case BLOCK_MOVE:
            if (help) Dest_Update();
		    HideBlock();
		    shrapnells->Hide();
			shrapnells->Update(ms);
			shrapnells->GetBackgnd();
			if (help) Dest_Show();
			ShowBlock();
			return ret;
		case BLOCK_STOP:
		    if (help) Dest_Update();
		    HideBlock();
		    shrapnells->Hide();
			InsertBlock();
			ShowBlock();
			ret = CheckLines(); //create new shrapnells
			DrawBlocks();
			shrapnells->Update(ms);
			shrapnells->GetBackgnd();
			if (ret == BOWL_RET_NONE)
				NextBlock();
			return ret;
			if (help) Dest_Show();
	}
	return ret;
}

void Bowl::FirstBlock()
{
	int i, j;
	//fill interior
	ClearBowl();
	if (adv_l_init) {
		//lines needed
		rem_lines = level_init[player->level+1].lines - level_init[player->level].lines;
		for (i = 0; i < BC_W; i++)
			for (j = 0; j < BC_H; j++)
				if (adv_l_init[player->level].bowlcontents[i][j] != 32)
					for (int a = 0; a < BLOCKSIZE; a++)
						bowl[i+BC_X][j*BLOCKSIZE + BLOCKSIZE*BC_Y + a] = adv_l_init[player->level].bowlcontents[i][j] - 96;
		DrawBlocks();
	}
	next = &block_init[rand() % block_num];
	NextBlock();
}

void Bowl::NextBlock()
{
	block->New(4, BLOCKSIZE, next, level_init[player->level].per_msec, hori_delay);
	BlockInit *old = next;
	next = &block_init[rand() % block_num];
	if (old == next)
		next = &block_init[rand() % block_num];
	new_next = 1;
	Dest_Update();
	dest_state = 0;
}

void Bowl::HideBlock()
{
	int old_x, old_y;
	int dx, dy;
	int *bstate;
	block->OldPos(&old_x, &old_y);
	bstate = block->OldState();
	if (!smooth)
    	old_x = (old_x - 1) * BLOCKSIZE + bowl_x;
    else	
        old_x = (int)block->OldCurX() - BLOCKSIZE + bowl_x;
	old_y += bowl_y;
	dx = old_x;
	dy = old_y;
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			if (dy > -BLOCKSIZE && ((int[4][4])bstate)[i][j])
				backgnd->Blit(dx, dy, screen, dx, dy, BLOCKSIZE, BLOCKSIZE);
			dy += BLOCKSIZE;
		}
		dx += BLOCKSIZE;
		dy = old_y;
	}
	if (*rect_num < DRAWRECTS) {
		SDL_Rect rect = {old_x, old_y, BLOCKSIZE*4, BLOCKSIZE*4};
		if (rect.y < 0) {
			rect.h += rect.y;
			rect.y = 0;
		}
		if (rect.x < 0) {
			rect.w += rect.x;
			rect.x = 0;
		}
		if (rect.y + rect.h >= screen->h)
			rect.h = screen->h - rect.y;
		if (rect.x + rect.w >= screen->w)
			rect.w = screen->w - rect.x;
		if (rect.h > 0 && rect.w > 0)	
    		rects[(*rect_num)++] = rect;
	}
}

void Bowl::ShowBlock()
{
	int x;
	if (!smooth)
	    x = (block->X() - 1) * BLOCKSIZE + bowl_x;
	else
    	x = (int)block->CurX() - BLOCKSIZE + bowl_x;
	int y = block->Y() + bowl_y;
	int dx, dy;
	int *bstate;
	bstate = block->State();
	dx = x;
	dy = y;
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			if (dy > -BLOCKSIZE && ((int[4][4])bstate)[i][j]) {
				if (trp)
					SC_Trp_X(screen->pixels, screen->w, dx, dy, blocks->Surface()->pixels, blocks->Width(), block->Id() * BLOCKSIZE, 0, BLOCKSIZE, BLOCKSIZE, BLOCK_TRP);
				else
					SC_Clp(screen->pixels, screen->w, dx, dy, blocks->Surface()->pixels, blocks->Width(), block->Id() * BLOCKSIZE, 0, BLOCKSIZE, BLOCKSIZE);
			}
			dy += BLOCKSIZE;
		}
		dx += BLOCKSIZE;
		dy = y;
	}
	block->SetOld();
	if (*rect_num < DRAWRECTS) {
		SDL_Rect rect = {x, y, BLOCKSIZE*4, BLOCKSIZE*4};
		if (rect.y < 0) {
			rect.h += rect.y;
			rect.y = 0;
		}
		if (rect.x < 0) {
			rect.w += rect.x;
			rect.x = 0;
		}
		if (rect.y + rect.h >= screen->h)
			rect.h = screen->h - rect.y;
		if (rect.x + rect.w >= screen->w)
			rect.w = screen->w - rect.x;
		if (rect.h > 0 && rect.w > 0)	
    		rects[(*rect_num)++] = rect;
	}
}

void Bowl::InsertBlock()
{
	int *bstate = block->State();
	int x, y;
	x = block->X();
	y = block->Y();
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			if (((int[4][4])bstate)[i][j])
				for (int b = 0; b < BLOCKSIZE; b++)
					bowl[x+i][y + j*BLOCKSIZE + b] = block->Id() + 1;
}

void Bowl::DrawBlocks()
{
    int scr_x = bowl_x < 0 ? 0 : bowl_x;
    int scr_y = bowl_y < 0 ? 0 : bowl_y;
	backgnd->Blit(scr_x, scr_y, screen, scr_x, scr_y, bowl_vis_w, bowl_vis_h);
	for (int i = 0; i < BOWL_W; i++)
		for (int j = 0; j < BOWL_H; j++)
			if (bowl[i + 1][j*BLOCKSIZE]) {
				if (trp)
					SC_Trp_X(screen->pixels, screen->w, bowl_x + i * BLOCKSIZE, bowl_y + j * BLOCKSIZE, blocks->Surface()->pixels, blocks->Width(), (bowl[i + 1][j * BLOCKSIZE] - 1) * BLOCKSIZE, 0, BLOCKSIZE, BLOCKSIZE, BLOCK_TRP);
				else
					SC_Clp(screen->pixels, screen->w, bowl_x + i * BLOCKSIZE, bowl_y + j * BLOCKSIZE, blocks->Surface()->pixels, blocks->Width(), (bowl[i + 1][j * BLOCKSIZE] - 1) * BLOCKSIZE, 0, BLOCKSIZE, BLOCKSIZE);
			}
	if (*rect_num < DRAWRECTS) {
		SDL_Rect	rect = {scr_x, scr_y, bowl_vis_w, bowl_vis_h};
		rects[(*rect_num)++] = rect;
	}
}

int Bowl::CheckLines()
{
    int enemy_has_lost;
    int ret = BOWL_RET_NONE;
    int i, j;
    int interior[BOWL_W][BOWL_H];
    int send_line_num;
    memset(interior, 0, sizeof(interior));
    //get interior
    for (i = 0; i < BOWL_W; i++)
        for (j = 0; j < BOWL_H; j++)
	        interior[i][j] = bowl[i+1][j*BLOCKSIZE];
	//check lines
	int lines = 0, c, line_y[4] = {-1, -1, -1, -1};
	for (j = 0; j < BOWL_H; j++) {
		c = 0;
		for (i = 0; i < BOWL_W; i++)
			if (interior[i][j])
				c++;
		if (c == BOWL_W)
			line_y[lines++] = j;
	}
	if (adv_l_init != 0 && lines > 0 && adv_l_init[player->level].dismantle_saves)
		last_action = 0;
	
	//create new shrapnells
	if (anim_l) {
		int s_lines[4] = {line_y[0]*BLOCKSIZE, line_y[1]*BLOCKSIZE, line_y[2]*BLOCKSIZE, line_y[3]*BLOCKSIZE};
		AddShrapnells(bowl_x, s_lines, lines);
	}
	
	// send lines
	enemy_has_lost = 0;
    if (lines && enemy) {
        send_line_num = send_all ? lines : (lines == 4 && send_tetris) ? lines : lines - 1;
        if (send_line_num && SendLines(interior, line_y, send_line_num))
            enemy_has_lost = BOWL_RET_ENEMYHASLOST;
    }

    //clear empty lines
    if (lines)
        for (j = 0; j < lines; j++)
	        for (i = 0; i < BOWL_W; i++)
		        interior[i][line_y[j]] = 0;

	//sum up
   	for (j = 0; j < lines; j++)
   		for (int k = 0; k < BOWL_W; k++)
   			for (int l = line_y[j]; l > 0; l--)
   				interior[k][l] = interior[k][l-1];
	//refresh bowl
	for (i = 0; i < BOWL_W; i++)
		for (j = 0; j < BOWL_H; j++)
			for (int k = 0; k < BLOCKSIZE; k++)
				bowl[i+1][j*BLOCKSIZE+k] = interior[i][j];

	//check action
	if (enemy == 0)
	    AdvAction();
			
	//use lines to set players score
	if (lines) {
#ifdef SOUND
		sndsrv->PlayWave(snd_explosion);
#endif
		player->lines += lines;
		if (total_lines)
		    (*total_lines) += lines;
		if (adv_l_init != 0) {
			rem_lines -= lines;
			if (rem_lines < 0)
				rem_lines = 0;
		}
		int base = 1;	
		while (lines--) {
		
		    if ( !init_data->slow )
    			player->score += base * level_init[player->level].score;
    		else
    			player->score += (base * level_init[player->level].score) / 2;
    			
			base++;
		}
	}
	if (!TooHigh()) {
	    if (total_lines) {
	        //multiplayer
	        if (player->level < level_num - 1 && *total_lines + line_offset >= level_init[player->level + 1].lines)
	            NextLevel();
	    }
	    else
		    if (player->level < level_num - 1 && player->lines + line_offset >= level_init[player->level+1].lines) {
			    if (adv_l_init != 0)
				    ret = BOWL_RET_NEWLEVEL | enemy_has_lost;
			    NextLevel();
		    }
		if (adv_l_init != 0 && player->level == LEVELNUMBER-1) {
			rem_lines = 0;
#ifdef SOUND
			sndsrv->PlayWave(snd_explosion);
#endif			
			return (BOWL_RET_GAMEOVER  | enemy_has_lost);
		}
		return ret;
	}
	else {
#ifdef SOUND
		sndsrv->PlayWave(snd_explosion);
#endif		
		return BOWL_RET_GAMEOVER;
	}
}

int Bowl::TooHigh()
{
	for (int i = 0; i < BOWL_W; i++)
		if (bowl[i + 1][BOWL_MIN_Y*BLOCKSIZE+1])
			return 1;
	return 0;
}

int Bowl::NewNextBlock()
{
	int ret = new_next;
	new_next = 0;
	return ret;
}

BlockInit* Bowl::Next()
{
	return next;
}

void Bowl::NextLevel()
{
	player->level++;
	if (adv_l_init == 0) return;
	//check height
	int i, j;
	for (j = BOWL_MIN_Y; j < BOWL_H ; j++) {
		for (i = 0; i < BOWL_W; i++)
			if (bowl[i + 1][j]) break;
		if (i != 10) break;
	}
	int base = 1;
	if (adv_l_init != 0) {
		int v = j - 4;
		while (v--) {
		
		    if ( !init_data->slow )
    			player->score += base * adv_l_init[player->level-1].line_score;
    		else
    			player->score += (base * adv_l_init[player->level-1].line_score) / 2;
    			
			base++;
		}
	}
	//lines needed
	rem_lines = level_init[player->level+1].lines - level_init[player->level].lines;
	//fill interior
	ClearBowl();
	for (i = 0; i < BC_W; i++)
		for (j = 0; j < BC_H; j++)
			if (adv_l_init[player->level].bowlcontents[i][j] != 32)
				for (int a = 0; a < BLOCKSIZE; a++)
					bowl[i+BC_X][j*BLOCKSIZE + BLOCKSIZE*BC_Y + a] = adv_l_init[player->level].bowlcontents[i][j] - 96;
	DrawBlocks();
}

void Bowl::AdvAction()
{
	if (adv_l_init == 0 ||
		adv_l_init[player->level].action == LA_NONE ||
		last_action < adv_l_init[player->level].delay) return;
	last_action = 0;
	//interior
	int i, j, c, pos;
	int interior[BOWL_W][BOWL_H];
	char line[BOWL_W];
	memset(line, 0, BOWL_W);
	memset(interior, 0, sizeof(interior));
	for (i = 0; i < BOWL_W; i++)
		for (j = 0; j < BOWL_H; j++)
			if (bowl[i+1][j*BLOCKSIZE])
				interior[i][j] = bowl[i+1][j*BLOCKSIZE];
	//action
	switch (adv_l_init[player->level].action) {
		case LA_ADDROW:
			for (i = 0; i < BOWL_W; i++)
				for (j = 0; j < BOWL_H - 1; j++)
					interior[i][j] = interior[i][j+1];
			for (i = 0; i < adv_l_init[player->level].strength; i++) {
				pos = rand() % BOWL_W;
				while (!line[pos]) {
					pos++;
					if (pos == BOWL_W) pos = 0;
				}
				line[pos] = 0;
			}
			for (i = 0; i < BOWL_W; i++)
				if (line[i])
					interior[i][BOWL_H - 1] = (rand() % block_num)+1;
				else
					interior[i][BOWL_H - 1] = 0;
			break;
		case LA_ADDBLOCK:
			pos = rand() % BOWL_W;
			i = BOWL_MIN_Y;
			while (!interior[pos][i])
				i++;
			i--;
			if (i < BOWL_MIN_Y)
				break;
			c = 0;
			for (j = 0; j < BOWL_W; j++)
				if (interior[j][i]) c++;
			if (c == BOWL_W - 1)
				pos++;
			if (pos == BOWL_W)
				pos -= 2;
			i = BOWL_MIN_Y;
			while (!interior[pos][i])
				i++;
			i--;
			if (i < BOWL_MIN_Y)
				break;
			interior[pos][i] = 10;
			break;
	}
	//redraw
	for (i = 0; i < BOWL_W; i++)
		for (j = 0; j < BOWL_H; j++)
			if (interior[i][j]) {
				for (int k = 0; k < BLOCKSIZE; k++)
					bowl[i+1][j*BLOCKSIZE+k] = interior[i][j];
			}
			else {
				for (int k = 0; k < BLOCKSIZE; k++)
					bowl[i+1][j*BLOCKSIZE+k] = 0;
			}
}

int* Bowl::RemLines()
{
	return &rem_lines;
}

void Bowl::AddShrapnells(int s_x, int *s_y, int ln)
{
	int i, j, a = 0, b = 0, v = 0;
	int add = 0, r_max = 0;
	Vector g = {0, 0.0001};
	Vector z = {0 ,0};
	Vector dir;
	float pct_ch = 0.04;
	int	r;
	if (anim_l == 1) //low: avoid slow animations...
	    r = rand() % 3 + 2;
	else
	    r = rand() % 5;
	int	all_lines[BOWL_H];
	if (s_y == 0) {
		r = 1;
		for (i = 0; i < BOWL_H; i++)
			all_lines[i] = i*BLOCKSIZE;
		s_y = all_lines;
		ln = BOWL_H;
	}
	switch (r) {
		case 1:
			switch (rand() % 3) {
				case 0:
					r_max = 20;
					add = 10;
					v = 1;
					break;
				case 1:
					r_max = 60;
					add = -30;
					v = 1;
					break;
				case 2:
					r_max = 20;
					add = 10;
					v = -1;
					break;
			}
			for (i = 0; i < ln; i++)
				for (j = 0; j < BOWL_W; j++)
					if (bowl[j+1][s_y[i]])
						for (a = 0;  a < 4; a++)
							for (b = 0; b < 4; b++) {
								dir.x = v * (float)(rand() % r_max + add) / 1000;
								dir.y = -(float)(rand() % 60 + 10) / 1000;
								shrapnells->Add(new Shrapnell(screen, s_x + j*BLOCKSIZE + a*4, s_y[i]+bowl_y+b*4, 4, 4, dir, g, pct_ch));
							}
			break;
		case 0:
			for (i = 0; i < ln; i++)
				for (j = 0; j < 5; j++)
					for (int a = 0; a < 8; a++){
						dir.x = -(float)(30 + j*16+a*2) / 1000;
						dir.y = 0;
						shrapnells->Add(new Shrapnell(screen, s_x + j*BLOCKSIZE + a*2, s_y[i]+bowl_y, 2, BLOCKSIZE, dir, z, pct_ch));
					}
			for (i = 0; i < ln; i++)
				for (j = 5; j < 10; j++)
					for (int a = 0; a < 8; a++){
						dir.x = (float)(30 + (9-j)*16+(7-a)*2) / 1000;
						dir.y = 0;
						shrapnells->Add(new Shrapnell(screen, s_x + j*BLOCKSIZE + a*2, s_y[i]+bowl_y, 2, BLOCKSIZE, dir, z, pct_ch));
					}
			break;
		case 2:
			for (i = 0; i < ln; i++)
				for (j = 0; j < 10; j++) {
					dir.x = 0;
					dir.y = (float)(50 + j*5) / 1000;
					shrapnells->Add(new Shrapnell(screen, s_x + j*BLOCKSIZE, s_y[i]+bowl_y, BLOCKSIZE, BLOCKSIZE, dir, z, pct_ch));
				}
			break;
		case 3:
			for (i = 0; i < ln; i++)
				for (j = 0; j < 10; j++) {
					dir.x = 0;
					dir.y = (float)(50 + (9-j)*5) / 1000;
					shrapnells->Add(new Shrapnell(screen, s_x + j*BLOCKSIZE, s_y[i]+bowl_y, BLOCKSIZE, BLOCKSIZE, dir, z, pct_ch));
				}
			break;
		case 4:
			for (i = 0; i < ln; i++)
				for (a = 0;  a < 8; a++) {
					dir.x = 0;
					dir.y = -(float)(30+a*10) / 1000;
					shrapnells->Add(new Shrapnell(screen, s_x, s_y[i]+bowl_y + a*2, BLOCKSIZE*10, 2, dir, g, pct_ch));
				}
			break;
	}
}

int Bowl::X()
{
	return bowl_x;
}

void Bowl::SetEnemy(Bowl *e)
{
    enemy = e;
}

int Bowl::SendLines(int interior[BOWL_W][BOWL_H], int *l_y, int num)
{
	int lines[4][BOWL_W];
	memset(lines, 0, BOWL_W * 4);
	for (int c = 0; c < num; c++)
	    for (int i = 0; i < BOWL_W; i++)
	        lines[c][i] = interior[i][l_y[c]];
	return enemy->ReceiveLines(lines, num);
}

int Bowl::ReceiveLines(int lines[4][BOWL_W], int num)
{
    int ret = 0;
    int i, j, k, pos;
    int interior[BOWL_W][BOWL_H];

    memset(interior, 0, sizeof(interior));

    //get interior
    for (i = 0; i < BOWL_W; i++)
        for (j = 0; j < BOWL_H; j++)
            interior[i][j] = bowl[i + 1][j * BLOCKSIZE];

    //raise lines
    for (j = 0; j < BOWL_H - num; j++)
        for (i = 0; i < BOWL_W; i++)
            interior[i][j] = interior[i][j + num];

    //check for game over
    for (i = 0; i < BOWL_W; i++)
        if (interior[i + 1][BOWL_MIN_Y - 1]) {
            ret = 1;
            break;
        }

    //add wholes in lines
    for (j = 0; j < num; j++) {
        for (k = 0; k < wholes; k++) {
            while (!lines[j][(pos = rand() % BOWL_W)]);
            lines[j][pos] = 0;
        }
    }

    //add lines
    for (i = 0; i < BOWL_W; i++)
        for (j = 0; j < num; j++)
            interior[i][BOWL_H - j - 1] = lines[j][i];

    //write lines to bowl
    for (i = 0; i < BOWL_W; i++)
        for (j = 0; j < BOWL_H; j++)
            for (k = 0; k < BLOCKSIZE; k++)
                bowl[i + 1][j * BLOCKSIZE + k] = interior[i][j];

    //update
    DrawBlocks();
    shrapnells->GetBackgnd();
    block->AdjustPos();
    ShowBlock();

    return ret;
}

void Bowl::Dest_Hide()
{
    int i, j;
    int x = (dest_x - 1) * BLOCKSIZE + bowl_x;
    int y = dest_y * BLOCKSIZE + bowl_y;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (block_init[block->Id()].blocks[dest_state][i][j])
                backgnd->Blit(x + i * BLOCKSIZE, y + j * BLOCKSIZE, screen, x + i * BLOCKSIZE, y + j * BLOCKSIZE, BLOCKSIZE, BLOCKSIZE);
    SDL_Rect rect = {x, y, BLOCKSIZE * 4, BLOCKSIZE * 4};
    if (rect.y < 0) {
        rect.h += rect.y;
        rect.y = 0;
    }
    if (rect.h + rect.y > screen->h)
        rect.h = screen->h - rect.y;
    if (rect.h > 0 && *rect_num < DRAWRECTS)
        rects[(*rect_num)++] = rect;
}

void Bowl::Dest_Show()
{
    int i, j;
    int x = (dest_x - 1) * BLOCKSIZE + bowl_x;
    int y = dest_y * BLOCKSIZE + bowl_y;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (block_init[block->Id()].blocks[block->Rot()][i][j])
                blocks->Blit(BLOCKSIZE * 10, 0, screen, x + i * BLOCKSIZE, y + j * BLOCKSIZE, BLOCKSIZE, BLOCKSIZE);
    SDL_Rect rect = {x, y, BLOCKSIZE * 4, BLOCKSIZE * 4};
    if (rect.y < 0) {
        rect.h += rect.y;
        rect.y = 0;
    }
    if (rect.h + rect.y > screen->h)
        rect.h = screen->h - rect.y;
    if (rect.h > 0 && *rect_num < DRAWRECTS)
        rects[(*rect_num)++] = rect;
}

void Bowl::Dest_Update()
{
    int i, j, l;

    dest_state = block->Rot();

    dest_x = block->X();
    for (l = block->Y(); l < BOWL_H * BLOCKSIZE; l += 16)
        for (i = 0; i < 4; i++)
            for (j = 3; j >= 0; j--)
                if (block_init[block->Id()].blocks[dest_state][i][j] && bowl[dest_x + i][l + j * BLOCKSIZE]) {
                    dest_y = l / BLOCKSIZE - BC_Y + 3;
                    return;
                }
}
