/***************************************************************************
                          sdlfont.cpp  -  description
                             -------------------
    begin                : Mon Feb 28 2000
    copyright            : (C) 2000 by Michael Speck
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "sdlfont.h"
#include "defs.h"

SDLFont::SDLFont(char *f)
{
    int i;

	data = new SDLSurface(f);
	if (data->Surface() == 0)
		exit(1);
	SetAlignment(TA_X_LEFT | TA_Y_TOP);
	SetColor(0x00FFFFFF);
	lh = data->Height();
	
	//table
	FILE	*file;
	file = fopen(f, "r");
	if (file == 0) {
    	char path[strlen(src_dir)+ strlen(f) + 1];
	    sprintf(path, "%s%s", src_dir, f);
	    file = fopen(path, "r");
	}
	fseek(file, -1, SEEK_END);
	fread(&offset, 1, 1, file);
	fseek(file, -2, SEEK_END);
	fread(&length, 1, 1, file);
	fseek(file, -2 - length, SEEK_END);
	fread(lw, 1, length, file);
	fclose(file);

	//letter offsets
	l_offset[0] = 0;
	for (i = 1; i < length; i++)	
		l_offset[i] = l_offset[i - 1] + lw[i - 1];
	
	//allowed keys
	memset(keys, 0, 256);
	for (i = 0; i < length; i++) {
		keys[i + offset] = 1;
	}
	
	lastX = lastY = lastW = lastH = 0;
}

SDLFont::SDLFont(char *f, int off, int len, int w)
{
    int i;

	data = new SDLSurface(f);
	if (data->Surface() == 0)
		exit(1);
	SetAlignment(TA_X_LEFT | TA_Y_TOP);
	SetColor(0x00FFFFFF);
	lh = data->Height();
	
	offset = off;
	length = len;
	
	for (i = 0; i < length; i++)
	    lw[i] = w;
	
	//letter offsets
	l_offset[0] = 0;
	for (i = 1; i < length; i++)	
		l_offset[i] = l_offset[i - 1] + w;
	
	//allowed keys
	memset(keys, 0, 256);
	for (i = 0; i < length; i++) {
		keys[i + offset] = 1;
	}
	
	lastX = lastY = lastW = lastH = 0;
}

SDLFont::~SDLFont()
{
	if (data) delete data;
}

void SDLFont::SetAlignment(int a)
{
	algn = a;
}

void SDLFont::SetColor(int c)
{
	color = c;
}

int  SDLFont::Color()
{
	return color;
}

int SDLFont::Write(SDL_Surface *dest, int x, int y, char *str, int pct)
{
	int	c_abs;
	int len = strlen(str);
	int pix_len = 0;
	int px = x, py = y;
	int i;
	
	pix_len = TextWidth(str);
	
	//alignment
	if (algn & TA_X_CENTER)
		px -= pix_len >> 1;
	else
		if (algn & TA_X_RIGHT)
			px -= pix_len;
	if (algn & TA_Y_CENTER)
		py -= lh >> 1;
	else
		if (algn & TA_Y_BOTTOM)
			py -= lh;

	SDL_Surface	*spf = SDL_GetVideoSurface();
	lastX = px; if (lastX < 0) lastX = 0;
	lastY = py; if (lastY < 0) lastY = 0;
	lastW = pix_len; if (lastX + lastW >= spf->w) lastW = spf->w - lastX;
	lastH = lh; if (lastY + lastH >= spf->h) lastH = spf->h - lastY;
			
	for (i = 0; i < len; i++) {
		c_abs = str[i] - offset;
		if (pct == 100)
			SC_Clp(dest->pixels, dest->w, px, py, data->Surface()->pixels, data->Width(), l_offset[c_abs], 0, lw[c_abs], lh);
		else
			SC_Trp_X(dest->pixels, dest->w, px, py, data->Surface()->pixels, data->Width(), l_offset[c_abs], 0, lw[c_abs], lh, pct);
		px += lw[c_abs];
	}
	
	return SDL_OK;
}

int SDLFont::Write(SDL_Surface *dest, int x, int y, char *str, int clr, int pct)
{
	int	c_abs;
	int len = strlen(str);
	int pix_len = 0;
	int px = x, py = y;
	int i;
	
	pix_len = TextWidth(str);
	
	//alignment
	if (algn & TA_X_CENTER)
		px -= pix_len >> 1;
	else
		if (algn & TA_X_RIGHT)
			px -= pix_len;
	if (algn & TA_Y_CENTER)
		py -= lh >> 1;
	else
		if (algn & TA_Y_BOTTOM)
			py -= lh;

	SDL_Surface	*spf = SDL_GetVideoSurface();
	lastX = px; if (lastX < 0) lastX = 0;
	lastY = py; if (lastY < 0) lastY = 0;
	lastW = pix_len; if (lastX + lastW >= spf->w) lastW = spf->w - lastX;
	lastH = lh; if (lastY + lastH >= spf->h) lastH = spf->h - lastY;
			
	for (i = 0; i < len; i++) {
		c_abs = str[i] - offset;
		SC_Clr_X(dest->pixels, dest->w, px, py, data->Surface()->pixels, data->Width(), l_offset[c_abs], 0, lw[c_abs], lh, clr, pct);
		px += lw[c_abs];
	}
	
	return SDL_OK;
}

void SDLFont::Begin()
{
	data->Begin();
}

void SDLFont::End()
{
	data->End();
}
	
int	SDLFont::LastX()
{
	return lastX;
}

int	SDLFont::LastY()
{
	return lastY;
}

int	SDLFont::LastW()
{
	return lastW;
}

int	SDLFont::LastH()
{
	return lastH;
}

SDL_Rect SDLFont::LastRect()
{
	SDL_Rect	rect={lastX, lastY, lastW, lastH};
	return rect;
}

int SDLFont::TextWidth(char *str)
{
	int pix_len = 0;
	for (unsigned int i = 0; i < strlen(str); i++)
		pix_len += lw[str[i] - offset];
	return pix_len;
}

int SDLFont::Height()
{
	return lh;
}
