/***************************************************************************
                          soundserver.cpp  -  description
                             -------------------
    begin                : Mon Mar 13 2000
    copyright            : (C) 2000 by Michael Speck
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <SDL/SDL.h>
#include <stdlib.h>
#include "soundserver.h"
#include "defs.h"

#ifdef SOUND

SoundServer::SoundServer(int t)
{
	int i;
	
	//empty tracks
	tracks = new Track[t];
	track_num = t;
	for (i = 0; i < t; i++)
		tracks[i].wave = 0;
	SetVolume(8);
	playing = 0;
	ok = 1;
}

SoundServer::~SoundServer()
{
	delete tracks;
}

int SoundServer::Open(SDL_AudioSpec wanted) {
	spec = wanted;
	spec.callback = SndSrv_Callback;
	spec.userdata = 0;
	spec.samples = 128;
	if (SDL_OpenAudio(&spec, 0) < 0) {
		printf("sndsrv: ERROR: open: %s\n", SDL_GetError());
		ok = 0;
		return RET_ERROR;
	}
	Pause(0);
	playing = 0;
	sleeping = 0;
	return RET_OK;
}

void SoundServer::Close()
{
    if (ok)
    	SDL_CloseAudio();
}

void SoundServer::Pause(int p)
{
    if (!ok) return;
	SDL_PauseAudio(p);
	if (p) playing = 0;
}

void SoundServer::PlayWave(Wave *w, int p = 0)
{
	if (sleeping || !ok) return;
	//use free track if possible
	for (int j = 0; j < track_num; j++)
		if (tracks[j].wave == 0) {
			//empty
			tracks[j].wave = w;
			tracks[j].audio_pos = w->Buffer();
			tracks[j].len = w->Length();
			playing++;
			Pause(0);
			return;
		}
	for (int j = 0; j < track_num; j++)
		if (tracks[j].priority <= p) {
			//lower priority found
			tracks[j].wave = w;
			tracks[j].audio_pos = w->Buffer();
			tracks[j].len = w->Length();
			playing++;
			Pause(0);
			return;
		}
	return; //cannot play sound, no track available
}

int	SoundServer::TrackNum()
{
	return track_num;
}

Track* SoundServer::GetTrack(int i)
{
    return &tracks[i];
}

char SoundServer::Volume()
{
	return volume;
}

void SoundServer::SetVolume(char v)
{
	volume = v < 0 ? 0 : v > 8 ? 8 : v;
	if (volume) volume = volume * 16 - 1;
}

int SoundServer::Playing()
{
	return playing;
}

void SoundServer::TrackFinished()
{
	playing--;
}

SDL_AudioSpec SoundServer::Spec()
{
	return spec;
}

void SoundServer::Activate(int a)
{
	sleeping = !a;
}

int SoundServer::Sleeping()
{
	return sleeping;
}

void SndSrv_Callback(void *udata, unsigned char *stream, int str_len)
{
	if (sndsrv->Sleeping()) return;
	int i, len;
	Track *track;
	for (i = 0; i < sndsrv->TrackNum(); i++) {
		track = sndsrv->GetTrack(i);
		if (track->wave == 0) continue;
		len = str_len < track->len ? str_len : track->len;
		SDL_MixAudio(stream, track->audio_pos, len, sndsrv->Volume());
		track->len -= len;
		track->audio_pos += len;
		if (track->len == 0) {
			track->wave = 0;
			sndsrv->TrackFinished();
		}
		if (!sndsrv->Playing()) sndsrv->Pause();
	}
}

#endif
